# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p10min_pb.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fp10min_pb.proto\x12\x0f\x61zlassets.proto\"+\n\x08\x63s_10800\x12\r\n\x05state\x18\x01 \x02(\r\x12\x10\n\x08platform\x18\x02 \x02(\t\"\xce\x01\n\x08sc_10801\x12\x12\n\ngateway_ip\x18\x01 \x02(\t\x12\x14\n\x0cgateway_port\x18\x02 \x02(\r\x12\x0b\n\x03url\x18\x03 \x02(\t\x12\x0f\n\x07version\x18\x04 \x03(\t\x12\x10\n\x08proxy_ip\x18\x05 \x01(\t\x12\x12\n\nproxy_port\x18\x06 \x01(\r\x12\r\n\x05is_ts\x18\x07 \x02(\r\x12\x11\n\ttimestamp\x18\x08 \x02(\r\x12 \n\x18monday_0oclock_timestamp\x18\t \x02(\r\x12\x10\n\x08\x63\x64n_list\x18\n \x03(\t')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'p10min_pb_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_CS_10800']._serialized_start=36
  _globals['_CS_10800']._serialized_end=79
  _globals['_SC_10801']._serialized_start=82
  _globals['_SC_10801']._serialized_end=288
# @@protoc_insertion_point(module_scope)
